@@ -50,6 +50,7 @@
 #include "verify.h"
 #include "display.h"
 #include "choose.h"
+#include "getvt.h"
 #include "gdm-net.h"
 
 /* Local functions */
@@ -68,7 +69,8 @@ static void gdm_restart_now (void);
 GSList *displays = NULL;	/* List of displays managed */
 GSList *xservers = NULL;	/* List of x server definitions */
 gint high_display_num = 0;	/* Highest local non-flexi display */
-gint sessions = 0;		/* Number of remote sessions */
+gint xdmcp_sessions = 0;	/* Number of remote sessions */
+gint xdmcp_pending = 0;		/* Number of pending remote sessions */
 gint flexi_servers = 0;		/* Number of flexi servers */
 uid_t GdmUserId;		/* Userid under which gdm should run */
 gid_t GdmGroupId;		/* Groupid under which gdm should run */
@@ -117,6 +119,7 @@ gboolean GdmSystemMenu = FALSE;
 gboolean GdmChooserButton = FALSE;
 gboolean GdmBrowser = FALSE;
 gboolean GdmAddGtkModules = FALSE;
+gboolean GdmDoubleLoginWarning = TRUE;
 gchar *GdmGlobalFaceDir = NULL;
 gint GdmXineramaScreen = 0;
 gchar *GdmGreeter = NULL;
@@ -178,7 +181,6 @@ gboolean GdmDisallowTCP = TRUE;
 /* set in the main function */
 char **stored_argv = NULL;
 int stored_argc = 0;
-char *stored_path = NULL;
 
 static time_t config_file_mtime = 0;
 
@@ -221,12 +223,12 @@ compare_displays (gconstpointer a, gconstpointer b)
 static void 
 gdm_config_parse (void)
 {
-    gchar *k, *v;
-    void *iter;
     struct passwd *pwent;
     struct group *grent;
     struct stat statbuf;
     gchar *bin;
+    VeConfig *cfg;
+    GList *list, *li;
     
     displays = NULL;
     high_display_num = 0;
@@ -239,42 +241,43 @@ gdm_config_parse (void)
     }
 
     /* Parse configuration options */
-    gnome_config_push_prefix ("=" GDM_CONFIG_FILE "=/");
+    cfg = ve_config_new (GDM_CONFIG_FILE);
 
-    GdmChooser = gnome_config_get_string (GDM_KEY_CHOOSER);
-    GdmDefaultPath = gnome_config_get_string (GDM_KEY_PATH);
-    GdmDisplayInit = gnome_config_get_string (GDM_KEY_INITDIR);
-    GdmAutomaticLoginEnable = gnome_config_get_bool (GDM_KEY_AUTOMATICLOGIN_ENABLE);
-    GdmAutomaticLogin = gnome_config_get_string (GDM_KEY_AUTOMATICLOGIN);
-    GdmAlwaysRestartServer = gnome_config_get_bool (GDM_KEY_ALWAYSRESTARTSERVER);
-    GdmGreeter = gnome_config_get_string (GDM_KEY_GREETER);
-    GdmRemoteGreeter = gnome_config_get_string (GDM_KEY_REMOTEGREETER);
-    GdmAddGtkModules = gnome_config_get_bool (GDM_KEY_ADD_GTK_MODULES);
-    GdmGtkModulesList = gnome_config_get_string (GDM_KEY_GTK_MODULES_LIST);	
-    GdmGroup = gnome_config_get_string (GDM_KEY_GROUP);
-    GdmHalt = gnome_config_get_string (GDM_KEY_HALT);
-    GdmKillInitClients = gnome_config_get_bool (GDM_KEY_KILLIC);
-    GdmLogDir= gnome_config_get_string (GDM_KEY_LOGDIR);
-    GdmPidFile = gnome_config_get_string (GDM_KEY_PIDFILE);
-    GdmSessionDir = gnome_config_get_string (GDM_KEY_SESSDIR);
-    GdmPostLogin = gnome_config_get_string (GDM_KEY_POSTLOGIN);
-    GdmPreSession = gnome_config_get_string (GDM_KEY_PRESESS);
-    GdmPostSession = gnome_config_get_string (GDM_KEY_POSTSESS);
-    GdmFailsafeXServer = gnome_config_get_string (GDM_KEY_FAILSAFE_XSERVER);
-    GdmXKeepsCrashing = gnome_config_get_string (GDM_KEY_XKEEPSCRASHING);
-    GdmConfigurator = gnome_config_get_string (GDM_KEY_CONFIGURATOR);
-    GdmConfigAvailable = gnome_config_get_bool (GDM_KEY_CONFIG_AVAILABLE);
-    GdmSystemMenu = gnome_config_get_bool (GDM_KEY_SYSMENU);
-    GdmChooserButton = gnome_config_get_bool (GDM_KEY_CHOOSER_BUTTON);
-    GdmBrowser = gnome_config_get_bool (GDM_KEY_BROWSER);
-    GdmGlobalFaceDir = gnome_config_get_string (GDM_KEY_FACEDIR);
-    GdmXineramaScreen = gnome_config_get_int (GDM_KEY_XINERAMASCREEN);
-    GdmReboot = gnome_config_get_string (GDM_KEY_REBOOT);
-    GdmRetryDelay = gnome_config_get_int (GDM_KEY_RETRYDELAY);
-    GdmRootPath = gnome_config_get_string (GDM_KEY_ROOTPATH);
-    GdmServAuthDir = gnome_config_get_string (GDM_KEY_SERVAUTH);
-    GdmSessDir = gnome_config_get_string (GDM_KEY_SESSDIR);
-    GdmXsession = gnome_config_get_string (GDM_KEY_BASEXSESSION);
+    GdmChooser = ve_config_get_string (cfg, GDM_KEY_CHOOSER);
+    GdmDefaultPath = ve_config_get_string (cfg, GDM_KEY_PATH);
+    GdmDisplayInit = ve_config_get_string (cfg, GDM_KEY_INITDIR);
+    GdmAutomaticLoginEnable = ve_config_get_bool (cfg, GDM_KEY_AUTOMATICLOGIN_ENABLE);
+    GdmAutomaticLogin = ve_config_get_string (cfg, GDM_KEY_AUTOMATICLOGIN);
+    GdmAlwaysRestartServer = ve_config_get_bool (cfg, GDM_KEY_ALWAYSRESTARTSERVER);
+    GdmGreeter = ve_config_get_string (cfg, GDM_KEY_GREETER);
+    GdmRemoteGreeter = ve_config_get_string (cfg, GDM_KEY_REMOTEGREETER);
+    GdmAddGtkModules = ve_config_get_bool (cfg, GDM_KEY_ADD_GTK_MODULES);
+    GdmDoubleLoginWarning = ve_config_get_bool (cfg, GDM_KEY_DOUBLELOGINWARNING);
+    GdmGtkModulesList = ve_config_get_string (cfg, GDM_KEY_GTK_MODULES_LIST);	
+    GdmGroup = ve_config_get_string (cfg, GDM_KEY_GROUP);
+    GdmHalt = ve_config_get_string (cfg, GDM_KEY_HALT);
+    GdmKillInitClients = ve_config_get_bool (cfg, GDM_KEY_KILLIC);
+    GdmLogDir= ve_config_get_string (cfg, GDM_KEY_LOGDIR);
+    GdmPidFile = ve_config_get_string (cfg, GDM_KEY_PIDFILE);
+    GdmSessionDir = ve_config_get_string (cfg, GDM_KEY_SESSDIR);
+    GdmPostLogin = ve_config_get_string (cfg, GDM_KEY_POSTLOGIN);
+    GdmPreSession = ve_config_get_string (cfg, GDM_KEY_PRESESS);
+    GdmPostSession = ve_config_get_string (cfg, GDM_KEY_POSTSESS);
+    GdmFailsafeXServer = ve_config_get_string (cfg, GDM_KEY_FAILSAFE_XSERVER);
+    GdmXKeepsCrashing = ve_config_get_string (cfg, GDM_KEY_XKEEPSCRASHING);
+    GdmConfigurator = ve_config_get_string (cfg, GDM_KEY_CONFIGURATOR);
+    GdmConfigAvailable = ve_config_get_bool (cfg, GDM_KEY_CONFIG_AVAILABLE);
+    GdmSystemMenu = ve_config_get_bool (cfg, GDM_KEY_SYSMENU);
+    GdmChooserButton = ve_config_get_bool (cfg, GDM_KEY_CHOOSER_BUTTON);
+    GdmBrowser = ve_config_get_bool (cfg, GDM_KEY_BROWSER);
+    GdmGlobalFaceDir = ve_config_get_string (cfg, GDM_KEY_FACEDIR);
+    GdmXineramaScreen = ve_config_get_int (cfg, GDM_KEY_XINERAMASCREEN);
+    GdmReboot = ve_config_get_string (cfg, GDM_KEY_REBOOT);
+    GdmRetryDelay = ve_config_get_int (cfg, GDM_KEY_RETRYDELAY);
+    GdmRootPath = ve_config_get_string (cfg, GDM_KEY_ROOTPATH);
+    GdmServAuthDir = ve_config_get_string (cfg, GDM_KEY_SERVAUTH);
+    GdmSessDir = ve_config_get_string (cfg, GDM_KEY_SESSDIR);
+    GdmXsession = ve_config_get_string (cfg, GDM_KEY_BASEXSESSION);
     if (ve_string_empty (GdmXsession)) {
 	    gdm_info (_("%s: BaseXsession empty, using %s/gdm/Xsession"),
 			"gdm_config_parse",
@@ -283,41 +286,41 @@ gdm_config_parse (void)
 	    GdmXsession = g_strconcat (EXPANDED_SYSCONFDIR, "/gdm/Xsession",
 				       NULL);
     }
-    GdmSuspend = gnome_config_get_string (GDM_KEY_SUSPEND);
-    GdmLocaleFile = gnome_config_get_string (GDM_KEY_LOCFILE);
+    GdmSuspend = ve_config_get_string (cfg, GDM_KEY_SUSPEND);
+    GdmLocaleFile = ve_config_get_string (cfg, GDM_KEY_LOCFILE);
 #if 0
 /* FIXME: Maybe just whack this */
-    GdmGnomeDefaultSession = gnome_config_get_string (GDM_KEY_GNOMEDEFAULTSESSION);
+    GdmGnomeDefaultSession = ve_config_get_string (cfg, GDM_KEY_GNOMEDEFAULTSESSION);
 #endif
-    GdmUser = gnome_config_get_string (GDM_KEY_USER);
-    GdmUserAuthDir = gnome_config_get_string (GDM_KEY_UAUTHDIR);
-    GdmUserAuthFile = gnome_config_get_string (GDM_KEY_UAUTHFILE);
-    GdmUserAuthFB = gnome_config_get_string (GDM_KEY_UAUTHFB);
+    GdmUser = ve_config_get_string (cfg, GDM_KEY_USER);
+    GdmUserAuthDir = ve_config_get_string (cfg, GDM_KEY_UAUTHDIR);
+    GdmUserAuthFile = ve_config_get_string (cfg, GDM_KEY_UAUTHFILE);
+    GdmUserAuthFB = ve_config_get_string (cfg, GDM_KEY_UAUTHFB);
 
-    GdmTimedLoginEnable = gnome_config_get_bool (GDM_KEY_TIMED_LOGIN_ENABLE);
-    GdmTimedLogin = gnome_config_get_string (GDM_KEY_TIMED_LOGIN);
-    GdmTimedLoginDelay = gnome_config_get_int (GDM_KEY_TIMED_LOGIN_DELAY);
+    GdmTimedLoginEnable = ve_config_get_bool (cfg, GDM_KEY_TIMED_LOGIN_ENABLE);
+    GdmTimedLogin = ve_config_get_string (cfg, GDM_KEY_TIMED_LOGIN);
+    GdmTimedLoginDelay = ve_config_get_int (cfg, GDM_KEY_TIMED_LOGIN_DELAY);
 
-    GdmAllowRoot = gnome_config_get_bool (GDM_KEY_ALLOWROOT);
-    GdmAllowRemoteRoot = gnome_config_get_bool (GDM_KEY_ALLOWREMOTEROOT);
-    GdmAllowRemoteAutoLogin = gnome_config_get_bool (GDM_KEY_ALLOWREMOTEAUTOLOGIN);
-    GdmRelaxPerms = gnome_config_get_int (GDM_KEY_RELAXPERM);
-    GdmUserMaxFile = gnome_config_get_int (GDM_KEY_MAXFILE);
-    GdmSessionMaxFile = gnome_config_get_int (GDM_KEY_SESSIONMAXFILE);
+    GdmAllowRoot = ve_config_get_bool (cfg, GDM_KEY_ALLOWROOT);
+    GdmAllowRemoteRoot = ve_config_get_bool (cfg, GDM_KEY_ALLOWREMOTEROOT);
+    GdmAllowRemoteAutoLogin = ve_config_get_bool (cfg, GDM_KEY_ALLOWREMOTEAUTOLOGIN);
+    GdmRelaxPerms = ve_config_get_int (cfg, GDM_KEY_RELAXPERM);
+    GdmUserMaxFile = ve_config_get_int (cfg, GDM_KEY_MAXFILE);
+    GdmSessionMaxFile = ve_config_get_int (cfg, GDM_KEY_SESSIONMAXFILE);
 
-    GdmXdmcp = gnome_config_get_bool (GDM_KEY_XDMCP);
-    GdmDispPerHost = gnome_config_get_int (GDM_KEY_DISPERHOST);
-    GdmMaxPending = gnome_config_get_int (GDM_KEY_MAXPEND);
-    GdmMaxManageWait = gnome_config_get_int (GDM_KEY_MAXWAIT);
-    GdmMaxSessions = gnome_config_get_int (GDM_KEY_MAXSESS);
-    GdmPort = gnome_config_get_int (GDM_KEY_UDPPORT);
-    GdmIndirect = gnome_config_get_bool (GDM_KEY_INDIRECT);
-    GdmMaxIndirect = gnome_config_get_int (GDM_KEY_MAXINDIR);
-    GdmMaxIndirectWait = gnome_config_get_int (GDM_KEY_MAXINDWAIT);    
-    GdmPingInterval = gnome_config_get_int (GDM_KEY_PINGINTERVAL);    
-    GdmWilling = gnome_config_get_string (GDM_KEY_WILLING);    
+    GdmXdmcp = ve_config_get_bool (cfg, GDM_KEY_XDMCP);
+    GdmDispPerHost = ve_config_get_int (cfg, GDM_KEY_DISPERHOST);
+    GdmMaxPending = ve_config_get_int (cfg, GDM_KEY_MAXPEND);
+    GdmMaxManageWait = ve_config_get_int (cfg, GDM_KEY_MAXWAIT);
+    GdmMaxSessions = ve_config_get_int (cfg, GDM_KEY_MAXSESS);
+    GdmPort = ve_config_get_int (cfg, GDM_KEY_UDPPORT);
+    GdmIndirect = ve_config_get_bool (cfg, GDM_KEY_INDIRECT);
+    GdmMaxIndirect = ve_config_get_int (cfg, GDM_KEY_MAXINDIR);
+    GdmMaxIndirectWait = ve_config_get_int (cfg, GDM_KEY_MAXINDWAIT);    
+    GdmPingInterval = ve_config_get_int (cfg, GDM_KEY_PINGINTERVAL);    
+    GdmWilling = ve_config_get_string (cfg, GDM_KEY_WILLING);    
 
-    GdmStandardXServer = gnome_config_get_string (GDM_KEY_STANDARD_XSERVER);    
+    GdmStandardXServer = ve_config_get_string (cfg, GDM_KEY_STANDARD_XSERVER);    
     if (ve_string_empty (GdmStandardXServer) ||
 	access (GdmStandardXServer, X_OK) != 0) {
 	    gdm_info (_("%s: Standard X server not found, trying alternatives"),
@@ -333,18 +336,16 @@ gdm_config_parse (void)
 		    GdmStandardXServer = g_strdup ("/usr/bin/X11/X");
 	    }
     }
-    GdmFlexibleXServers = gnome_config_get_int (GDM_KEY_FLEXIBLE_XSERVERS);    
-    GdmXnest = gnome_config_get_string (GDM_KEY_XNEST);    
+    GdmFlexibleXServers = ve_config_get_int (cfg, GDM_KEY_FLEXIBLE_XSERVERS);    
+    GdmXnest = ve_config_get_string (cfg, GDM_KEY_XNEST);    
     if (ve_string_empty (GdmXnest))
 	    GdmXnest = NULL;
 
-    GdmFirstVT = gnome_config_get_int (GDM_KEY_FIRSTVT);    
-    GdmVTAllocation = gnome_config_get_bool (GDM_KEY_VTALLOCATION);    
-    GdmDisallowTCP = gnome_config_get_bool (GDM_KEY_DISALLOWTCP);    
+    GdmFirstVT = ve_config_get_int (cfg, GDM_KEY_FIRSTVT);    
+    GdmVTAllocation = ve_config_get_bool (cfg, GDM_KEY_VTALLOCATION);    
+    GdmDisallowTCP = ve_config_get_bool (cfg, GDM_KEY_DISALLOWTCP);    
 
-    GdmDebug = gnome_config_get_bool (GDM_KEY_DEBUG);
-
-    gnome_config_pop_prefix();
+    GdmDebug = ve_config_get_bool (cfg, GDM_KEY_DEBUG);
 
     /* sanitize some values */
 
@@ -412,29 +413,26 @@ gdm_config_parse (void)
 	gdm_error (_("%s: No sessions directory specified."), "gdm_config_parse");
 
     /* Find server definitions */
-    iter = gnome_config_init_iterator_sections ("=" GDM_CONFIG_FILE "=/");
-    iter = gnome_config_iterator_next (iter, &k, NULL);
-    
-    while (iter) {
-	    if (strncmp (k, "server-", strlen ("server-")) == 0) {
-		    char *section;
+    list = ve_config_get_sections (cfg);
+    for (li = list; li != NULL; li = li->next) {
+	    const char *sec = li->data;
+	    if (strncmp (sec, "server-", strlen ("server-")) == 0) {
 		    GdmXServer *svr = g_new0 (GdmXServer, 1);
+		    char buf[256];
 
-		    section = g_strdup_printf ("=" GDM_CONFIG_FILE "=/%s/", k);
-		    gnome_config_push_prefix (section);
-
-		    svr->id = g_strdup (k + strlen ("server-"));
-		    svr->name = gnome_config_get_string (GDM_KEY_SERVER_NAME);
-		    svr->command = gnome_config_get_string
-			    (GDM_KEY_SERVER_COMMAND);
-		    svr->flexible = gnome_config_get_bool
-			    (GDM_KEY_SERVER_FLEXIBLE);
-		    svr->choosable = gnome_config_get_bool
-			    (GDM_KEY_SERVER_CHOOSABLE);
-		    svr->handled = gnome_config_get_bool
-			    (GDM_KEY_SERVER_HANDLED);
-		    svr->chooser = gnome_config_get_bool
-			    (GDM_KEY_SERVER_CHOOSER);
+		    svr->id = g_strdup (sec + strlen ("server-"));
+		    g_snprintf (buf, sizeof (buf), "%s/" GDM_KEY_SERVER_NAME, sec);
+		    svr->name = ve_config_get_string (cfg, buf);
+		    g_snprintf (buf, sizeof (buf), "%s/" GDM_KEY_SERVER_COMMAND, sec);
+		    svr->command = ve_config_get_string (cfg, buf);
+		    g_snprintf (buf, sizeof (buf), "%s/" GDM_KEY_SERVER_FLEXIBLE, sec);
+		    svr->flexible = ve_config_get_bool (cfg, buf);
+		    g_snprintf (buf, sizeof (buf), "%s/" GDM_KEY_SERVER_CHOOSABLE, sec);
+		    svr->choosable = ve_config_get_bool (cfg, buf);
+		    g_snprintf (buf, sizeof (buf), "%s/" GDM_KEY_SERVER_HANDLED, sec);
+		    svr->handled = ve_config_get_bool (cfg, buf);
+		    g_snprintf (buf, sizeof (buf), "%s/" GDM_KEY_SERVER_CHOOSER, sec);
+		    svr->chooser = ve_config_get_bool (cfg, buf);
 
 		    if (ve_string_empty (svr->command)) {
 			    gdm_error (_("%s: Empty server command, "
@@ -444,16 +442,10 @@ gdm_config_parse (void)
 			    svr->command = g_strdup (GdmStandardXServer);
 		    }
 
-		    g_free (section);
-		    gnome_config_pop_prefix ();
-
 		    xservers = g_slist_append (xservers, svr);
 	    }
-
-	    g_free (k);
-
-	    iter = gnome_config_iterator_next (iter, &k, NULL);
     }
+    ve_config_free_list_of_strings (list);
 
     if (xservers == NULL ||
 	gdm_find_x_server (GDM_STANDARD) == NULL) {
@@ -470,30 +462,33 @@ gdm_config_parse (void)
     }
 
     /* Find local X server definitions */
-    iter = gnome_config_init_iterator ("=" GDM_CONFIG_FILE "=/" GDM_KEY_SERVERS);
-    iter = gnome_config_iterator_next (iter, &k, &v);
-    
-    while (iter) {
-	    if (isdigit (*k)) {
-		    int disp_num = atoi (k);
+    list = ve_config_get_keys (cfg, GDM_KEY_SERVERS);
+    for (li = list; li != NULL; li = li->next) {
+	    const char *key = li->data;
+	    if (isdigit (*key)) {
+		    char *full;
+		    char *val;
+		    int disp_num = atoi (key);
 		    GdmDisplay *disp;
 
 		    while (display_exists (disp_num)) {
 			    disp_num++;
 		    }
 
-		    if (disp_num != atoi (k)) {
+		    if (disp_num != atoi (key)) {
 			    gdm_error (_("%s: Display number %d in use!  I will use %d"),
-				       "gdm_config_parse", atoi (k), disp_num);
+				       "gdm_config_parse", atoi (key), disp_num);
 		    }
 
-		    disp = gdm_server_alloc (disp_num, v);
-		    if (disp == NULL) {
-			    g_free (k);
-			    g_free (v);
-			    iter = gnome_config_iterator_next (iter, &k, &v);
+		    full = g_strdup_printf ("%s/%s", GDM_KEY_SERVERS, key);
+		    val = ve_config_get_string (cfg, full);
+		    g_free (full);
+
+		    disp = gdm_server_alloc (disp_num, val);
+		    g_free (val);
+
+		    if (disp == NULL)
 			    continue;
-		    }
 		    displays = g_slist_insert_sorted (displays,
 						      disp,
 						      compare_displays);
@@ -502,11 +497,8 @@ gdm_config_parse (void)
 	    } else {
 		    gdm_info (_("%s: Invalid server line in config file. Ignoring!"), "gdm_config_parse");
 	    }
-	    g_free (k);
-	    g_free (v);
-
-	    iter = gnome_config_iterator_next (iter, &k, &v);
     }
+    ve_config_free_list_of_strings (list);
 
     if (displays == NULL && ! GdmXdmcp) {
 	    char *server = NULL;
@@ -719,6 +711,8 @@ gdm_config_parse (void)
 
     /* Check that user authentication is properly configured */
     gdm_verify_check ();
+
+    ve_config_destroy (cfg);
 }
 
 /* If id == NULL, then get the first X server */
@@ -756,7 +750,7 @@ gdm_daemonify (void)
     pid = fork ();
     if (pid > 0) {
 
-	if ((pf = fopen (GdmPidFile, "w")) != NULL) {
+	if ((pf = gdm_safe_fopen_w (GdmPidFile)) != NULL) {
 	    fprintf (pf, "%d\n", (int)pid);
 	    fclose (pf);
 	}
@@ -866,7 +860,8 @@ gdm_final_cleanup (void)
 
 	/* Close stuff */
 
-	gdm_xdmcp_close ();
+	if (GdmXdmcp)
+		gdm_xdmcp_close ();
 
 	if (fifoconn != NULL) {
 		char *path;
@@ -941,21 +936,23 @@ deal_with_x_crashes (GdmDisplay *d)
 		    argv[0] = GdmXKeepsCrashing;
 		    argv[1] = NULL;
 
+		    gdm_restoreenv ();
+
 		    /* unset DISPLAY and XAUTHORITY if they exist
 		     * so that gdialog (if used) doesn't get confused */
-		    gnome_unsetenv ("DISPLAY");
-		    gnome_unsetenv ("XAUTHORITY");
+		    ve_unsetenv ("DISPLAY");
+		    ve_unsetenv ("XAUTHORITY");
 
 		    /* some promised variables */
-		    gnome_setenv ("XLOG", xlog, TRUE);
-		    gnome_setenv ("BINDIR", EXPANDED_BINDIR, TRUE);
-		    gnome_setenv ("SBINDIR", EXPANDED_SBINDIR, TRUE);
-		    gnome_setenv ("LIBEXECDIR", EXPANDED_LIBEXECDIR, TRUE);
-		    gnome_setenv ("SYSCONFDIR", EXPANDED_SYSCONFDIR, TRUE);
+		    ve_setenv ("XLOG", xlog, TRUE);
+		    ve_setenv ("BINDIR", EXPANDED_BINDIR, TRUE);
+		    ve_setenv ("SBINDIR", EXPANDED_SBINDIR, TRUE);
+		    ve_setenv ("LIBEXECDIR", EXPANDED_LIBEXECDIR, TRUE);
+		    ve_setenv ("SYSCONFDIR", EXPANDED_SYSCONFDIR, TRUE);
 
 		    /* To enable gettext stuff in the script */
-		    gnome_setenv ("TEXTDOMAIN", GETTEXT_PACKAGE, TRUE);
-		    gnome_setenv ("TEXTDOMAINDIR", GNOMELOCALEDIR, TRUE);
+		    ve_setenv ("TEXTDOMAIN", GETTEXT_PACKAGE, TRUE);
+		    ve_setenv ("TEXTDOMAINDIR", GNOMELOCALEDIR, TRUE);
 
 		    execv (argv[0], argv);
 	
@@ -1017,6 +1014,8 @@ suspend_machine (void)
 	if (GdmSuspendReal != NULL &&
 	    fork () == 0) {
 		char **argv;
+		if (GdmXdmcp)
+			gdm_xdmcp_close ();
 		/* In the child setup empty mask and set all signals to
 		 * default values */
 		gdm_unset_signals ();
@@ -1037,13 +1036,7 @@ suspend_machine (void)
 static void
 change_to_first_and_clear (gboolean reboot)
 {
-	/* kind of hack, we could do it ourselves if we weren't lazy */
-	if (access ("/usr/bin/chvt", X_OK) == 0)
-		system ("/usr/bin/chvt 1");
-	else if (access ("/bin/chvt", X_OK) == 0)
-		system ("/bin/chvt 1");
-	else
-		return;
+	gdm_change_vt (1);
 	close (1);
 	close (2);
 	open ("/dev/tty1", O_WRONLY);
@@ -1174,6 +1167,18 @@ gdm_cleanup_children (void)
 	    status = DISPLAY_REMANAGE;
     }
 
+    if (status == DISPLAY_GREETERFAILED) {
+	    if (d->managetime + 10 >= time (NULL)) {
+		    d->try_different_greeter = TRUE;
+	    } else {
+		    d->try_different_greeter = FALSE;
+	    }
+	    /* now just remanage */
+	    status = DISPLAY_REMANAGE;
+    } else {
+	    d->try_different_greeter = FALSE;
+    }
+
     /* checkout if we can actually do stuff */
     switch (status) {
     case DISPLAY_REBOOT:
@@ -1216,7 +1221,9 @@ start_autopsy:
 	gdm_final_cleanup ();
 	chdir ("/");
 
+#ifdef __linux__
 	change_to_first_and_clear (TRUE /* reboot */);
+#endif /* __linux */
 
 	argv = ve_split (GdmRebootReal);
 	execv (argv[0], argv);
@@ -1234,7 +1241,9 @@ start_autopsy:
 	gdm_final_cleanup ();
 	chdir ("/");
 
+#ifdef __linux__
 	change_to_first_and_clear (FALSE /* reboot */);
+#endif /* __linux */
 
 	argv = ve_split (GdmHaltReal);
 	execv (argv[0], argv);
@@ -1352,8 +1361,6 @@ start_autopsy:
     if (gdm_restart_mode)
 	    gdm_safe_restart ();
 
-    g_main_loop_quit (main_loop);
-
     return TRUE;
 }
 
@@ -1362,8 +1369,7 @@ gdm_restart_now (void)
 {
 	gdm_info (_("GDM restarting ..."));
 	gdm_final_cleanup ();
-	if (stored_path != NULL)
-		gnome_setenv ("PATH", stored_path, TRUE);
+	gdm_restoreenv ();
 	execvp (stored_argv[0], stored_argv);
 	gdm_error (_("Failed to restart self"));
 	_exit (1);
@@ -1427,8 +1433,6 @@ store_argv (int argc, char *argv[])
 {
 	int i;
 
-	stored_path = g_strdup (g_getenv ("PATH"));
-
 	stored_argv = g_new0 (char *, argc + 1);
 	for (i = 0; i < argc; i++)
 		stored_argv[i] = g_strdup (argv[i]);
@@ -1617,7 +1621,7 @@ main (int argc, char *argv[])
     if (no_daemon || getppid() == 1) {
 
 	/* Write pid to pidfile */
-	if ((pf = fopen (GdmPidFile, "w")) != NULL) {
+	if ((pf = gdm_safe_fopen_w (GdmPidFile)) != NULL) {
 	    fprintf (pf, "%d\n", (int)getpid());
 	    fclose (pf);
 	}
@@ -1741,7 +1745,7 @@ write_x_servers (GdmDisplay *d)
 	if (d->x_servers_order < 0)
 		d->x_servers_order = get_new_order (d);
 
-	fp = fopen (file, "w");
+	fp = gdm_safe_fopen_w (file);
 	if (fp == NULL) {
 		gdm_error ("Can't open %s for writing", file);
 		g_free (file);
@@ -1774,13 +1778,21 @@ write_x_servers (GdmDisplay *d)
 }
 
 static void
-send_slave_ack (GdmDisplay *d)
+send_slave_ack (GdmDisplay *d, const char *resp)
 {
 	if (d->master_notify_fd >= 0) {
-		char not[2];
-		not[0] = GDM_SLAVE_NOTIFY_ACK;
-		not[1] = '\n';
-		write (d->master_notify_fd, not, 2);
+		if (resp == NULL) {
+			char not[2];
+			not[0] = GDM_SLAVE_NOTIFY_ACK;
+			not[1] = '\n';
+			write (d->master_notify_fd, not, 2);
+		} else {
+			char *not = g_strdup_printf ("%c%s\n",
+						     GDM_SLAVE_NOTIFY_ACK,
+						     resp);
+			write (d->master_notify_fd, not, strlen (not));
+			g_free (not);
+		}
 	}
 	gdm_sigchld_block_push ();
 	if (d->slavepid > 1) {
@@ -1855,7 +1867,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->chosen_hostname = g_strdup (p);
 			gdm_debug ("Got CHOSEN_LOCAL == %s", p);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_XPID " ",
 		            strlen (GDM_SOP_XPID " ")) == 0) {
@@ -1873,7 +1885,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->servpid = pid;
 			gdm_debug ("Got XPID == %ld", (long)pid);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_SESSPID " ",
 		            strlen (GDM_SOP_SESSPID " ")) == 0) {
@@ -1891,7 +1903,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->sesspid = pid;
 			gdm_debug ("Got SESSPID == %ld", (long)pid);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_GREETPID " ",
 		            strlen (GDM_SOP_GREETPID " ")) == 0) {
@@ -1909,7 +1921,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->greetpid = pid;
 			gdm_debug ("Got GREETPID == %ld", (long)pid);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_CHOOSERPID " ",
 		            strlen (GDM_SOP_CHOOSERPID " ")) == 0) {
@@ -1927,7 +1939,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->chooserpid = pid;
 			gdm_debug ("Got CHOOSERPID == %ld", (long)pid);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_LOGGED_IN " ",
 		            strlen (GDM_SOP_LOGGED_IN " ")) == 0) {
@@ -1953,7 +1965,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 				gdm_safe_restart ();
 
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_DISP_NUM " ",
 		            strlen (GDM_SOP_DISP_NUM " ")) == 0) {
@@ -1974,7 +1986,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->dispnum = disp_num;
 			gdm_debug ("Got DISP_NUM == %d", disp_num);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_VT_NUM " ",
 		            strlen (GDM_SOP_VT_NUM " ")) == 0) {
@@ -1993,7 +2005,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->vt = vt_num;
 			gdm_debug ("Got VT_NUM == %d", vt_num);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_LOGIN " ",
 		            strlen (GDM_SOP_LOGIN " ")) == 0) {
@@ -2020,7 +2032,53 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->login = g_strdup (p);
 			gdm_debug ("Got LOGIN == %s", p);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
+		}
+	} else if (strncmp (msg, GDM_SOP_QUERYLOGIN " ",
+		            strlen (GDM_SOP_QUERYLOGIN " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+		char *p;
+
+		if (sscanf (msg, GDM_SOP_QUERYLOGIN " %ld",
+			    &slave_pid) != 1)
+			return;
+		p = strchr (msg, ' ');
+		if (p != NULL)
+			p = strchr (p+1, ' ');
+		if (p == NULL)
+			return;
+
+		p++;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			GString *resp = NULL;
+			GSList *li;
+			gdm_debug ("Got QUERYLOGIN %s", p);
+			for (li = displays; li != NULL; li = li->next) {
+				GdmDisplay *di = li->data;
+				if (di->login != NULL &&
+				    strcmp (di->login, p) == 0) {
+					if (resp == NULL) {
+						resp = g_string_new (di->name);
+						g_string_append_printf (resp, ",%d", di->vt);
+					} else {
+						g_string_append_printf (resp, ",%s,%d",
+									di->name, di->vt);
+					}
+				}
+			}
+
+			/* send ack */
+			if (resp != NULL) {
+				send_slave_ack (d, resp->str);
+				g_string_free (resp, TRUE);
+			} else {
+				send_slave_ack (d, NULL);
+			}
 		}
 	} else if (strncmp (msg, GDM_SOP_COOKIE " ",
 		            strlen (GDM_SOP_COOKIE " ")) == 0) {
@@ -2047,7 +2105,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			d->cookie = g_strdup (p);
 			gdm_debug ("Got COOKIE == <secret>");
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_FLEXI_ERR " ",
 		            strlen (GDM_SOP_FLEXI_ERR " ")) == 0) {
@@ -2084,7 +2142,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			
 			gdm_debug ("Got FLEXI_ERR == %d", err);
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strncmp (msg, GDM_SOP_FLEXI_OK " ",
 		            strlen (GDM_SOP_FLEXI_OK " ")) == 0) {
@@ -2114,7 +2172,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			
 			gdm_debug ("Got FLEXI_OK");
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strcmp (msg, GDM_SOP_SOFT_RESTART) == 0) {
 		gdm_restart_mode = TRUE;
@@ -2183,7 +2241,7 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			write_x_servers (d);
 
 			/* send ack */
-			send_slave_ack (d);
+			send_slave_ack (d, NULL);
 		}
 	} else if (strcmp (msg, GDM_SOP_SUSPEND_MACHINE) == 0) {
 		gdm_info (_("Master suspending..."));
@@ -2586,6 +2644,8 @@ static gboolean
 update_config (const char *key)
 {
 	struct stat statbuf;
+	VeConfig *cfg;
+
 	if (stat (GDM_CONFIG_FILE, &statbuf) == -1) {
 		/* if the file didn't exist before either */
 		if (config_file_mtime == 0)
@@ -2597,10 +2657,10 @@ update_config (const char *key)
 		config_file_mtime = statbuf.st_mtime;
 	}
 
-	gnome_config_push_prefix ("=" GDM_CONFIG_FILE "=/");
+	cfg = ve_config_new (GDM_CONFIG_FILE);
 
 	if (is_key (key, GDM_KEY_ALLOWROOT)) {
-		gboolean val = gnome_config_get_bool (GDM_KEY_ALLOWROOT);
+		gboolean val = ve_config_get_bool (cfg, GDM_KEY_ALLOWROOT);
 		if (ve_bool_equal (val, GdmAllowRoot))
 			goto update_config_ok;
 		GdmAllowRoot = val;
@@ -2609,7 +2669,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_ALLOWREMOTEROOT)) {
-		gboolean val = gnome_config_get_bool (GDM_KEY_ALLOWREMOTEROOT);
+		gboolean val = ve_config_get_bool (cfg, GDM_KEY_ALLOWREMOTEROOT);
 		if (ve_bool_equal (val, GdmAllowRemoteRoot))
 			goto update_config_ok;
 		GdmAllowRemoteRoot = val;
@@ -2618,7 +2678,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_ALLOWREMOTEAUTOLOGIN)) {
-		gboolean val = gnome_config_get_bool (GDM_KEY_ALLOWREMOTEAUTOLOGIN);
+		gboolean val = ve_config_get_bool (cfg, GDM_KEY_ALLOWREMOTEAUTOLOGIN);
 		if (ve_bool_equal (val, GdmAllowRemoteAutoLogin))
 			goto update_config_ok;
 		GdmAllowRemoteAutoLogin = val;
@@ -2627,7 +2687,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_SYSMENU)) {
-		gboolean val = gnome_config_get_bool (GDM_KEY_SYSMENU);
+		gboolean val = ve_config_get_bool (cfg, GDM_KEY_SYSMENU);
 		if (ve_bool_equal (val, GdmSystemMenu))
 			goto update_config_ok;
 		GdmSystemMenu = val;
@@ -2636,7 +2696,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_CONFIG_AVAILABLE)) {
-		gboolean val = gnome_config_get_bool (GDM_KEY_CONFIG_AVAILABLE);
+		gboolean val = ve_config_get_bool (cfg, GDM_KEY_CONFIG_AVAILABLE);
 		if (ve_bool_equal (val, GdmConfigAvailable))
 			goto update_config_ok;
 		GdmConfigAvailable = val;
@@ -2645,7 +2705,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_CHOOSER_BUTTON)) {
-		gboolean val = gnome_config_get_bool (GDM_KEY_CHOOSER_BUTTON);
+		gboolean val = ve_config_get_bool (cfg, GDM_KEY_CHOOSER_BUTTON);
 		if (ve_bool_equal (val, GdmChooserButton))
 			goto update_config_ok;
 		GdmChooserButton = val;
@@ -2654,7 +2714,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_RETRYDELAY)) {
-		int val = gnome_config_get_int (GDM_KEY_RETRYDELAY);
+		int val = ve_config_get_int (cfg, GDM_KEY_RETRYDELAY);
 		if (val == GdmRetryDelay)
 			goto update_config_ok;
 		GdmRetryDelay = val;
@@ -2663,7 +2723,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_DISALLOWTCP)) {
-		gboolean val = gnome_config_get_bool (GDM_KEY_DISALLOWTCP);
+		gboolean val = ve_config_get_bool (cfg, GDM_KEY_DISALLOWTCP);
 		if (ve_bool_equal (val, GdmDisallowTCP))
 			goto update_config_ok;
 		GdmDisallowTCP = val;
@@ -2672,7 +2732,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_GREETER)) {
-		char *val = gnome_config_get_string (GDM_KEY_GREETER);
+		char *val = ve_config_get_string (cfg, GDM_KEY_GREETER);
 		if (strcmp (ve_sure_string (val), ve_sure_string (GdmGreeter)) == 0) {
 			g_free (val);
 			goto update_config_ok;
@@ -2684,7 +2744,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_REMOTEGREETER)) {
-		char *val = gnome_config_get_string (GDM_KEY_REMOTEGREETER);
+		char *val = ve_config_get_string (cfg, GDM_KEY_REMOTEGREETER);
 		if (strcmp (ve_sure_string (val), ve_sure_string (GdmRemoteGreeter)) == 0) {
 			g_free (val);
 			goto update_config_ok;
@@ -2697,14 +2757,14 @@ update_config (const char *key)
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_TIMED_LOGIN) ||
 		   is_key (key, GDM_KEY_TIMED_LOGIN_ENABLE)) {
-		gboolean enable = gnome_config_get_bool (GDM_KEY_TIMED_LOGIN_ENABLE);
+		gboolean enable = ve_config_get_bool (cfg, GDM_KEY_TIMED_LOGIN_ENABLE);
 		char *val;
 
 		/* if not enabled, we just don't care */
 		if ( ! enable && ! GdmTimedLoginEnable)
 			goto update_config_ok;
 
-		val = gnome_config_get_string (GDM_KEY_TIMED_LOGIN);
+		val = ve_config_get_string (cfg, GDM_KEY_TIMED_LOGIN);
 		if (strcmp (ve_sure_string (val),
 			    ve_sure_string (GdmTimedLogin)) == 0 &&
 		    ve_bool_equal (enable, GdmTimedLoginEnable)) {
@@ -2726,7 +2786,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_TIMED_LOGIN_DELAY)) {
-		int val = gnome_config_get_int (GDM_KEY_TIMED_LOGIN_DELAY);
+		int val = ve_config_get_int (cfg, GDM_KEY_TIMED_LOGIN_DELAY);
 		if (val == GdmTimedLoginDelay)
 			goto update_config_ok;
 		GdmTimedLoginDelay = val;
@@ -2735,7 +2795,7 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, GDM_KEY_XDMCP)) {
-		gboolean val = gnome_config_get_bool (GDM_KEY_XDMCP);
+		gboolean val = ve_config_get_bool (cfg, GDM_KEY_XDMCP);
 		if (ve_bool_equal (val, GdmXdmcp))
 			goto update_config_ok;
 		GdmXdmcp = val;
@@ -2749,16 +2809,16 @@ update_config (const char *key)
 
 		goto update_config_ok;
 	} else if (is_key (key, "xdmcp/PARAMETERS")) {
-		GdmDispPerHost = gnome_config_get_int (GDM_KEY_DISPERHOST);
-		GdmMaxPending = gnome_config_get_int (GDM_KEY_MAXPEND);
-		GdmMaxManageWait = gnome_config_get_int (GDM_KEY_MAXWAIT);
-		GdmMaxSessions = gnome_config_get_int (GDM_KEY_MAXSESS);
-		GdmIndirect = gnome_config_get_bool (GDM_KEY_INDIRECT);
-		GdmMaxIndirect = gnome_config_get_int (GDM_KEY_MAXINDIR);
-		GdmMaxIndirectWait = gnome_config_get_int (GDM_KEY_MAXINDWAIT);
-		GdmPingInterval = gnome_config_get_int (GDM_KEY_PINGINTERVAL);
+		GdmDispPerHost = ve_config_get_int (cfg, GDM_KEY_DISPERHOST);
+		GdmMaxPending = ve_config_get_int (cfg, GDM_KEY_MAXPEND);
+		GdmMaxManageWait = ve_config_get_int (cfg, GDM_KEY_MAXWAIT);
+		GdmMaxSessions = ve_config_get_int (cfg, GDM_KEY_MAXSESS);
+		GdmIndirect = ve_config_get_bool (cfg, GDM_KEY_INDIRECT);
+		GdmMaxIndirect = ve_config_get_int (cfg, GDM_KEY_MAXINDIR);
+		GdmMaxIndirectWait = ve_config_get_int (cfg, GDM_KEY_MAXINDWAIT);
+		GdmPingInterval = ve_config_get_int (cfg, GDM_KEY_PINGINTERVAL);
 	} else if (is_key (key, GDM_KEY_UDPPORT)) {
-		int val = gnome_config_get_int (GDM_KEY_UDPPORT);
+		int val = ve_config_get_int (cfg, GDM_KEY_UDPPORT);
 		if (GdmPort == val)
 			goto update_config_ok;
 		GdmPort = val;
@@ -2770,12 +2830,12 @@ update_config (const char *key)
 		}
 	}
 
-	gnome_config_pop_prefix ();
+	ve_config_destroy (cfg);
 	return FALSE;
 
 update_config_ok:
 
-	gnome_config_pop_prefix ();
+	ve_config_destroy (cfg);
 	return TRUE;
 }
 
@@ -2935,6 +2995,20 @@ gdm_handle_user_message (GdmConnection *conn, const char *msg, gpointer data)
 			}
 		}
 		gdm_connection_write (conn, "\n");
+	} else if (strcmp (msg, GDM_SUP_ALL_SERVERS) == 0) {
+		GSList *li;
+		const char *sep = " ";
+		gdm_connection_write (conn, "OK");
+		for (li = displays; li != NULL; li = li->next) {
+			GdmDisplay *disp = li->data;
+			gdm_connection_printf (conn,
+					       "%s%s,%s",
+					       sep,
+					       ve_sure_string (disp->name),
+					       ve_sure_string (disp->login));
+			sep = ";";
+		}
+		gdm_connection_write (conn, "\n");
 	} else if (strcmp (msg, GDM_SUP_GREETERPIDS) == 0) {
 		GSList *li;
 		const char *sep = " ";
