@@ -54,14 +54,14 @@
 #include "misc.h"
 #include "gdmwm.h"
 
-/* set the DOING_GDM_DEVELOPMENT env variable if you want to
- * search for the glade file in the current dir and not the system
- * install dir, better then something you have to change
- * in the source and recompile */
-static gboolean DOING_GDM_DEVELOPMENT = FALSE;
-
 static gboolean RUNNING_UNDER_GDM = FALSE;
 
+enum {
+	CHOOSER_LIST_ICON_COLUMN = 0,
+	CHOOSER_LIST_LABEL_COLUMN,
+	CHOOSER_LIST_HOST_COLUMN
+};
+
 typedef struct _GdmChooserHost GdmChooserHost;
 struct _GdmChooserHost {
     gchar *name;
@@ -85,17 +85,25 @@ static void gdm_chooser_abort (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 static void gdm_chooser_warn (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 
 /* Exported for glade */
-gboolean gdm_chooser_cancel (void);
-gboolean gdm_chooser_manage (GtkButton *button, gpointer data);
-gboolean gdm_chooser_browser_select (GtkWidget *widget,
-				     gint selected,
-				     GdkEvent *event);
-gboolean gdm_chooser_browser_unselect (GtkWidget *widget,
-				       gint selected,
-				       GdkEvent *event);
-gboolean gdm_chooser_xdmcp_discover (void);
+void gdm_chooser_cancel (void);
+void gdm_chooser_add_host (void);
+void gdm_chooser_manage (GtkButton *button, gpointer data);
+void gdm_chooser_browser_select (GtkWidget *widget,
+				 gint selected,
+				 GdkEvent *event);
+void gdm_chooser_browser_unselect (GtkWidget *widget,
+				   gint selected,
+				   GdkEvent *event);
+void gdm_chooser_xdmcp_discover (void);
 void display_chooser_information (void);
 
+#define ADD_TIMEOUT 3000
+static guint add_check_handler = 0;
+
+/* if this is received, select that host automatically */
+static struct in_addr *added_addr = NULL;
+static char *added_host = NULL;
+
 static guint scan_time_handler = 0;
 
 #define PING_TIMEOUT 2000
@@ -134,6 +142,7 @@ static gchar *GdmHostDefaultIcon;
 static gchar *GdmGtkRC;
 static gchar *GdmHosts;
 static gboolean GdmBroadcast;
+static gboolean GdmAllowAdd;
 static gchar *GdmBackgroundColor;
 static int GdmBackgroundType;
 enum {
@@ -143,15 +152,32 @@ enum {
 };
 
 static GladeXML *chooser_app;
-static GtkWidget *chooser, *manage, *rescan, *cancel;
+static GtkWidget *chooser, *manage, *rescan, *cancel, *add_entry;
 static GtkWidget *status_label;
 
 static GIOChannel *channel;
 static GList *hosts = NULL;
 static GdkPixbuf *defhostimg;
 static GtkWidget *browser;
+static GtkTreeModel *browser_model;
 static GdmChooserHost *curhost;
 
+static gboolean
+find_host_in_list (GdmChooserHost *host, GtkTreeIter *iter)
+{
+	if (gtk_tree_model_get_iter_first (browser_model, iter)) {
+		do {
+			GdmChooserHost *lhost;
+			gtk_tree_model_get (browser_model, iter,
+					    CHOOSER_LIST_HOST_COLUMN, &lhost,
+					    -1);
+			if (lhost == host)
+				return TRUE;
+		} while (gtk_tree_model_iter_next (browser_model, iter));
+	}
+	return FALSE;
+}
+
 static void
 setup_cursor (GdkCursorType type)
 {
@@ -177,18 +203,6 @@ gdm_chooser_host_dispose (GdmChooserHost *host)
     g_free (host);
 }
 
-static gint 
-gdm_chooser_sort_func (gpointer d1, gpointer d2)
-{
-    GdmChooserHost *a = d1;
-    GdmChooserHost *b = d2;
-
-    if (!d1 || !d2)
-	return 0;
-
-    return strcmp (a->name, b->name);
-}
-
 static GdmChooserHost * 
 gdm_chooser_host_alloc (const char *hostname,
 			const char *description,
@@ -198,7 +212,6 @@ gdm_chooser_host_alloc (const char *hostname,
     GdmChooserHost *host;
     GdkPixbuf *img;
     gchar *hostimg;
-    GList *hostl;
 
     host = g_new0 (GdmChooserHost, 1);
     host->name = g_strdup (hostname);
@@ -206,19 +219,7 @@ gdm_chooser_host_alloc (const char *hostname,
     memcpy (&host->ia, ia, sizeof (struct in_addr));
     host->willing = willing;
 
-    hostl = g_list_find_custom (hosts,
-				host,
-				(GCompareFunc) gdm_chooser_sort_func);
-    /* replace */
-    if (hostl != NULL) {
-	    GdmChooserHost *old = hostl->data;
-	    hostl->data = host;
-	    gdm_chooser_host_dispose (old);
-    } else {
-	    hosts = g_list_insert_sorted (hosts, 
-					  host,
-					  (GCompareFunc) gdm_chooser_sort_func);
-    }
+    hosts = g_list_prepend (hosts, host);
     
     if ( ! willing)
 	    return host;
@@ -263,46 +264,56 @@ gdm_chooser_host_alloc (const char *hostname,
 }
 
 static void
-gdm_chooser_browser_update (void)
+gdm_chooser_browser_add_host (GdmChooserHost *host)
 {
-    GList *li;
-    gboolean any;
+    if (host->willing) {
+	    GtkTreeIter iter = {0};
+	    const char *addr = inet_ntoa (host->ia);
+	    char *label;
 
-    gnome_icon_list_freeze (GNOME_ICON_LIST (browser));
-    gnome_icon_list_clear (GNOME_ICON_LIST (browser));
+	    if (strcmp (addr, host->name) == 0)
+		    label = g_strdup_printf ("%s\n%s",
+					     host->name,
+					     host->desc);
+	    else
+		    label = g_strdup_printf ("%s (%s)\n%s",
+					     host->name,
+					     addr,
+					     host->desc);
 
-    any = FALSE;
-    for (li = hosts; li != NULL; li = li->next) {
-	    GdmChooserHost *host = (GdmChooserHost *) li->data;
+	    gtk_list_store_append (GTK_LIST_STORE (browser_model), &iter);
+	    gtk_list_store_set (GTK_LIST_STORE (browser_model), &iter,
+				CHOOSER_LIST_ICON_COLUMN, host->picture,
+				CHOOSER_LIST_LABEL_COLUMN, label,
+				CHOOSER_LIST_HOST_COLUMN, host,
+				-1);
+	    g_free (label);
 
-	    if (host->willing) {
-		    /* FIXME: the \n doesn't actually propagate
-		     * since the icon list is a broken piece of horsedung */
-		    char *temp = g_strconcat (host->name, " \n",
-					      host->desc, NULL);
-		    gnome_icon_list_append_pixbuf (GNOME_ICON_LIST (browser),
-						   host->picture,
-						   NULL /* icon_filename */,
-						   temp);
-		    g_free (temp);
-		    any = TRUE;
+	    if (added_addr != NULL &&
+		memcmp (&host->ia, added_addr,
+			sizeof (struct in_addr)) == 0) {
+		    GtkTreeSelection *selection;
+		    GtkTreePath *path = gtk_tree_model_get_path (browser_model, &iter);
+		    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (browser));
+		    gtk_tree_selection_select_iter (selection, &iter);
+		    gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (browser),
+						  path, NULL,
+						  FALSE, 0.0, 0.0);
+		    gtk_tree_path_free (path);
+		    gtk_widget_grab_focus (manage);
 	    }
+	    g_free (added_host);
+	    added_host = NULL;
+	    added_addr = NULL;
+	    if (add_check_handler > 0)
+		    g_source_remove (add_check_handler);
+	    add_check_handler = 0;
     }
 
-    gnome_icon_list_thaw (GNOME_ICON_LIST (browser));
-
-    if (any) {
-      gtk_label_set_text (GTK_LABEL (status_label), _(active_network));
-    } else {
-      gtk_label_set_text (GTK_LABEL (status_label), _(empty_network));
-    }
-    gtk_widget_set_sensitive (GTK_WIDGET (manage), FALSE);
-    gtk_widget_set_sensitive (GTK_WIDGET (rescan), TRUE);
     gtk_widget_set_sensitive (GTK_WIDGET (browser), TRUE);
 }
 
-
-static gboolean
+static GdmChooserHost *
 gdm_host_known (struct in_addr *ia)
 {
 	GList *li;
@@ -310,10 +321,10 @@ gdm_host_known (struct in_addr *ia)
 	for (li = hosts; li != NULL; li = li->next) {
 		GdmChooserHost *host = li->data;
 		if (memcmp (&host->ia, ia, sizeof (struct in_addr)) == 0) {
-			return TRUE;
+			return host;
 		}
 	}
-	return FALSE;
+	return NULL;
 }
 
 static gboolean
@@ -329,6 +340,20 @@ is_loopback_addr (struct in_addr *ia)
 	}
 }
 
+static struct in_addr *
+gdm_addr_known (struct in_addr *ia)
+{
+	GSList *li;
+
+	for (li = queryaddr; li != NULL; li = li->next) {
+		struct in_addr *a = li->data;
+		if (memcmp (a, ia, sizeof (struct in_addr)) == 0) {
+			return a;
+		}
+	}
+	return NULL;
+}
+
 static gboolean
 gdm_chooser_decode_packet (GIOChannel   *source,
 			   GIOCondition  condition,
@@ -342,6 +367,7 @@ gdm_chooser_decode_packet (GIOChannel   *source,
     gchar *hostname = NULL;
     gchar *status = NULL;
     ARRAY8 auth, host, stat;
+    GdmChooserHost *gh;
 
     if (! XdmcpFill (sockfd, &buf, (XdmcpNetaddr) &clnt_sa, &sa_len))
 	return TRUE;
@@ -390,14 +416,62 @@ gdm_chooser_decode_packet (GIOChannel   *source,
 	    }
     }
 
-    gdm_chooser_host_alloc (hostname,
-			    status,
-			    &clnt_sa.sin_addr,
-			    header.opcode == WILLING);
+    gh = gdm_host_known (&clnt_sa.sin_addr);
+    if (gh == NULL) {
+	    gh = gdm_chooser_host_alloc (hostname,
+					 status,
+					 &clnt_sa.sin_addr,
+					 header.opcode == WILLING);
+
+
+	    gdm_chooser_browser_add_host (gh);
+    } else {
+	    /* server changed it's mind */
+	    if (header.opcode == WILLING &&
+		! gh->willing) {
+		    gh->willing = TRUE;
+		    gdm_chooser_browser_add_host (gh);
+	    }
+	    /* hmmm what about the other change, just ignore
+	       for now, it's kind of confusing to just remove
+	       servers really */
+    }
+
+    if ( ! gh->willing &&
+	added_addr != NULL &&
+	memcmp (&gh->ia, added_addr,
+		sizeof (struct in_addr)) == 0) {
+	    GtkWidget *dialog;
+
+	    added_addr = NULL;
+	    if (add_check_handler > 0)
+		    g_source_remove (add_check_handler);
+	    add_check_handler = 0;
+
+	    dialog = ve_hig_dialog_new
+		    (GTK_WINDOW (chooser) /* parent */,
+		     GTK_DIALOG_MODAL /* flags */,
+		     GTK_MESSAGE_ERROR,
+		     GTK_BUTTONS_OK,
+		     _("Cannot connect to remote server"),
+		     _("The host \"%s\" is not willing "
+		       "to support a login session right now.  "
+		       "Please try again later."),
+		     added_host);
+
+	    if (RUNNING_UNDER_GDM)
+		    gdm_wm_center_window (GTK_WINDOW (dialog));
+
+	    gdm_wm_no_login_focus_push ();
+	    gtk_dialog_run (GTK_DIALOG (dialog));
+	    gtk_widget_destroy (dialog);
+	    gdm_wm_no_login_focus_pop ();
+
+	    g_free (added_host);
+	    added_host = NULL;
+    }
 
     g_free (hostname);
-
-    gdm_chooser_browser_update ();
     
  done:
     if (header.opcode == WILLING) {
@@ -476,8 +550,19 @@ gdm_chooser_find_bcaddr (void)
 static gboolean
 chooser_scan_time_update (gpointer data)
 {
+	GList *li;
 	scan_time_handler = 0;
-	gdm_chooser_browser_update ();
+	for (li = hosts; li != NULL; li = li->next) {
+		GdmChooserHost *host = (GdmChooserHost *) li->data;
+		if (host->willing)
+			break;
+	}
+	if (li != NULL /* something was found */) {
+		gtk_label_set_text (GTK_LABEL (status_label), _(active_network));
+	} else {
+		gtk_label_set_text (GTK_LABEL (status_label), _(empty_network));
+	}
+	gtk_widget_set_sensitive (GTK_WIDGET (rescan), TRUE);
 	return FALSE;
 }
 
@@ -522,13 +607,21 @@ ping_try (gpointer data)
 		return TRUE;
 }
 
-gboolean 
+void
 gdm_chooser_xdmcp_discover (void)
 {
     GList *hl = hosts;
 
+    g_free (added_host);
+    added_host = NULL;
+    added_addr = NULL;
+    if (add_check_handler > 0)
+	    g_source_remove (add_check_handler);
+    add_check_handler = 0;
+
     gtk_widget_set_sensitive (GTK_WIDGET (manage), FALSE);
-    gnome_icon_list_clear (GNOME_ICON_LIST (browser));
+    gtk_widget_set_sensitive (GTK_WIDGET (rescan), FALSE);
+    gtk_list_store_clear (GTK_LIST_STORE (browser_model));
     gtk_widget_set_sensitive (GTK_WIDGET (browser), FALSE);
     gtk_label_set_text (GTK_LABEL (status_label),
 			_(scanning_message));
@@ -554,8 +647,6 @@ gdm_chooser_xdmcp_discover (void)
     if (ping_try_handler > 0)
 	    g_source_remove (ping_try_handler);
     ping_try_handler = g_timeout_add (PING_TIMEOUT, ping_try, NULL);
-
-    return TRUE;
 }
 
 #ifndef ishexdigit
@@ -718,18 +809,163 @@ gdm_chooser_choose_host (const char *hostname)
   }
 }
 
+static gboolean
+add_check (gpointer data)
+{
+	if (added_addr != NULL) {
+		GtkWidget *dialog;
 
-gboolean
+		dialog = ve_hig_dialog_new
+			(GTK_WINDOW (chooser) /* parent */,
+			 GTK_DIALOG_MODAL /* flags */,
+			 GTK_MESSAGE_ERROR,
+			 GTK_BUTTONS_OK,
+			 _("Did not receive response from server"),
+			 _("Did not receive any response from host \"%s\" "
+			   "in %d seconds.  Perhaps the host is not "
+			   "turned on, or is not willing to support a "
+			   "login session right now.  Please try again "
+			   "later."),
+			 added_host,
+			 ADD_TIMEOUT / 1000);
+
+		if (RUNNING_UNDER_GDM)
+			gdm_wm_center_window (GTK_WINDOW (dialog));
+
+		gdm_wm_no_login_focus_push ();
+		gtk_dialog_run (GTK_DIALOG (dialog));
+		gtk_widget_destroy (dialog);
+		gdm_wm_no_login_focus_pop ();
+	}
+	add_check_handler = 0;
+	return FALSE;
+}
+
+void
+gdm_chooser_add_host (void)
+{
+	struct hostent *hostent;
+	struct sockaddr_in qa;
+	struct in_addr *ia;
+	GdmChooserHost *host;
+	struct sockaddr_in sock;
+	const char *name;
+
+	name = gtk_entry_get_text (GTK_ENTRY (add_entry));
+	if (ve_string_empty (name))
+		return;
+
+	if (strlen (name) == 8 &&
+	    from_hex (name, (char *) &qa.sin_addr, strlen (name)) == 0)
+		qa.sin_family = AF_INET;
+	else if ((qa.sin_addr.s_addr = inet_addr (name)) != -1)
+		qa.sin_family = AF_INET;
+	else if ((hostent = gethostbyname (name)) != NULL
+		 && hostent->h_addrtype == AF_INET
+		 && hostent->h_length == 4) {
+		qa.sin_family = AF_INET;
+		memmove (&qa.sin_addr, hostent->h_addr, 4);
+	} else {
+		GtkWidget *dialog;
+
+		dialog = ve_hig_dialog_new
+			(GTK_WINDOW (chooser) /* parent */,
+			 GTK_DIALOG_MODAL /* flags */,
+			 GTK_MESSAGE_ERROR,
+			 GTK_BUTTONS_OK,
+			 _("Cannot find host"),
+			 _("I cannot find the host \"%s\", "
+			   "perhaps you have mistyped it."),
+			 name);
+
+		if (RUNNING_UNDER_GDM)
+			gdm_wm_center_window (GTK_WINDOW (dialog));
+
+		gdm_wm_no_login_focus_push ();
+		gtk_dialog_run (GTK_DIALOG (dialog));
+		gtk_widget_destroy (dialog);
+		gdm_wm_no_login_focus_pop ();
+		return; /* not a valid address */
+	}
+
+	sock.sin_family = AF_INET;
+	sock.sin_port = htons (XDM_UDP_PORT);
+
+	ia = gdm_addr_known (&qa.sin_addr);
+	if (ia == NULL) {
+		ia = g_new0 (struct in_addr, 1);
+		ia->s_addr = qa.sin_addr.s_addr;
+		queryaddr = g_slist_append (queryaddr, ia);
+	}
+
+	host = gdm_host_known (ia);
+	if (host != NULL) {
+		GtkTreeIter iter = {0};
+		if (find_host_in_list (host, &iter)) {
+			GtkTreeSelection *selection;
+			GtkTreePath *path = gtk_tree_model_get_path (browser_model, &iter);
+			selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (browser));
+			gtk_tree_selection_select_iter (selection, &iter);
+			gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (browser),
+						      path, NULL,
+						      FALSE, 0.0, 0.0);
+			gtk_tree_path_free (path);
+			gtk_widget_grab_focus (manage);
+		} else {
+			/* hmm, probably not willing, ping the host then for
+			   good measure */
+			sock.sin_addr.s_addr = ia->s_addr;
+			XdmcpFlush (sockfd, &querybuf, (XdmcpNetaddr) &sock,
+				    (int)sizeof (struct sockaddr_in));
+
+			added_addr = ia;
+			g_free (added_host);
+			added_host = g_strdup (name);
+
+			if (add_check_handler > 0)
+				g_source_remove (add_check_handler);
+			add_check_handler = g_timeout_add (ADD_TIMEOUT,
+							   add_check, NULL);
+		}
+
+		/* empty the text entry to indicate success */
+		gtk_entry_set_text (GTK_ENTRY (add_entry), "");
+		return;
+	}
+
+	added_addr = ia;
+	g_free (added_host);
+	added_host = g_strdup (name);
+
+	/* and send out the query */
+	sock.sin_addr.s_addr = ia->s_addr;
+	XdmcpFlush (sockfd, &querybuf, (XdmcpNetaddr) &sock, (int)sizeof (struct sockaddr_in));
+
+	if (add_check_handler > 0)
+		g_source_remove (add_check_handler);
+	add_check_handler = g_timeout_add (ADD_TIMEOUT,
+					   add_check, NULL);
+
+	/* empty the text entry to indicate success */
+	gtk_entry_set_text (GTK_ENTRY (add_entry), "");
+}
+
+void
 gdm_chooser_cancel (void)
 {
-    closelog();
-    gtk_main_quit();
+    if (scan_time_handler > 0) {
+	    g_source_remove (scan_time_handler);
+	    scan_time_handler = 0;
+    }
 
-    return TRUE;
+    closelog ();
+    /* exit rather gtk_main_quit, it's just safer this way we don't
+       have to worry about random whackiness happening */
+    exit (EXIT_SUCCESS);
 }
 
 
-gboolean
+void
 gdm_chooser_manage (GtkButton *button, gpointer data)
 {
     if (scan_time_handler > 0) {
@@ -741,9 +977,10 @@ gdm_chooser_manage (GtkButton *button, gpointer data)
       gdm_chooser_choose_host (curhost->name);
    
     closelog();
-    gtk_main_quit();
-   
-    return TRUE;
+
+    /* exit rather gtk_main_quit, it's just safer this way we don't
+       have to worry about random whackiness happening */
+    exit (EXIT_SUCCESS);
 }
 
 
@@ -781,29 +1018,27 @@ gdm_chooser_warn (const gchar *format, ...)
 static void 
 gdm_chooser_parse_config (void)
 {
-    /* stupid, or is it?
-    struct stat unused;
-    if (stat (GDM_CONFIG_FILE, &unused) == -1)
-	gdm_chooser_abort (_("gdm_chooser_parse_config: No configuration file: %s. Aborting."), GDM_CONFIG_FILE);
-	*/
+    VeConfig *cfg;
 
-    gnome_config_push_prefix ("=" GDM_CONFIG_FILE "=/");
+    cfg = ve_config_get (GDM_CONFIG_FILE);
 
-    GdmXineramaScreen = gnome_config_get_int (GDM_KEY_XINERAMASCREEN);
-    GdmGtkRC = gnome_config_get_string (GDM_KEY_GTKRC);
-    GdmScanTime = gnome_config_get_int (GDM_KEY_SCAN);
-    GdmHostDefaultIcon = gnome_config_get_string (GDM_KEY_HOST);
-    GdmHostIconDir = gnome_config_get_string (GDM_KEY_HOSTDIR);
-    GdmIconMaxWidth = gnome_config_get_int (GDM_KEY_ICONWIDTH);
-    GdmIconMaxHeight = gnome_config_get_int (GDM_KEY_ICONHEIGHT);
-    GdmDebug = gnome_config_get_bool (GDM_KEY_DEBUG);
+    GdmXineramaScreen = ve_config_get_int (cfg, GDM_KEY_XINERAMASCREEN);
+    GdmGtkRC = ve_config_get_string (cfg, GDM_KEY_GTKRC);
+    GdmScanTime = ve_config_get_int (cfg, GDM_KEY_SCAN);
+    GdmHostDefaultIcon = ve_config_get_string (cfg, GDM_KEY_HOST);
+    GdmHostIconDir = ve_config_get_string (cfg, GDM_KEY_HOSTDIR);
+    GdmIconMaxWidth = ve_config_get_int (cfg, GDM_KEY_ICONWIDTH);
+    GdmIconMaxHeight = ve_config_get_int (cfg, GDM_KEY_ICONHEIGHT);
+    GdmDebug = ve_config_get_bool (cfg, GDM_KEY_DEBUG);
 
-    GdmBackgroundColor = gnome_config_get_string (GDM_KEY_BACKGROUNDCOLOR);
-    GdmBackgroundType = gnome_config_get_int (GDM_KEY_BACKGROUNDTYPE);
+    GdmBackgroundColor = ve_config_get_string (cfg, GDM_KEY_BACKGROUNDCOLOR);
+    GdmBackgroundType = ve_config_get_int (cfg, GDM_KEY_BACKGROUNDTYPE);
+
+    GdmAllowAdd = ve_config_get_bool (cfg, GDM_KEY_ALLOWADD);
 
     /* note that command line arguments will prevail over these */
-    GdmHosts = gnome_config_get_string (GDM_KEY_HOSTS);
-    GdmBroadcast = gnome_config_get_bool (GDM_KEY_BROADCAST);
+    GdmHosts = ve_config_get_string (cfg, GDM_KEY_HOSTS);
+    GdmBroadcast = ve_config_get_bool (cfg, GDM_KEY_BROADCAST);
     /* if broadcasting, then append BROADCAST to hosts */
     if (GdmBroadcast) {
 	    if (ve_string_empty (GdmHosts)) {
@@ -819,75 +1054,31 @@ gdm_chooser_parse_config (void)
     if (GdmScanTime < 1) GdmScanTime = 1;
     if (GdmIconMaxWidth < 0) GdmIconMaxWidth = 128;
     if (GdmIconMaxHeight < 0) GdmIconMaxHeight = 128;
-
-    gnome_config_pop_prefix();
 }
 
-static GdmChooserHost *
-gdm_nth_willing_host (int n)
+static void
+host_selected (GtkTreeSelection *selection, gpointer data)
 {
-	GList *li;
-	int i;
+	GtkTreeModel *tm = NULL;
+	GtkTreeIter iter = {0};
 
-	i = 0;
-	for (li = hosts; li != NULL; li = li->next) {
-		GdmChooserHost *host = li->data;
-		if (host->willing) {
-			if (i == n)
-				return host;
-			i++;
-		}
-	}
-	return NULL;
-}
-
-
-gboolean 
-gdm_chooser_browser_select (GtkWidget *widget, gint selected, GdkEvent *event)
-{
-    if (!widget || !event)
-	return TRUE;
-
-    switch (event->type) {
-	
-    case GDK_BUTTON_PRESS:
-    case GDK_BUTTON_RELEASE:
-	curhost = gdm_nth_willing_host (selected);
-	gtk_widget_set_sensitive (manage, TRUE);
-	break;
-
-    case GDK_2BUTTON_PRESS:
-	curhost = gdm_nth_willing_host (selected);
-	gdm_chooser_manage (NULL, NULL);
-	break;
-	
-    default: 
-	break;
-    }
-    
-    return TRUE;
-}
-
-
-gboolean
-gdm_chooser_browser_unselect (GtkWidget *widget, gint selected, GdkEvent *event)
-{
-    if (!event) 
-	return TRUE;
-
-    switch (event->type) {
-	
-    case GDK_BUTTON_PRESS:
-    case GDK_BUTTON_RELEASE:
 	curhost = NULL;
-	gtk_widget_set_sensitive (manage, FALSE);
-	break;
-	
-    default:
-	break;
-    }
 
-    return TRUE;
+	if (gtk_tree_selection_get_selected (selection, &tm, &iter)) {
+		gtk_tree_model_get (tm, &iter, CHOOSER_LIST_HOST_COLUMN,
+				    &curhost, -1);
+	}
+
+	gtk_widget_set_sensitive (manage, curhost != NULL);
+}
+
+static void
+row_activated (GtkTreeView *tree_view,
+	       GtkTreePath *path,
+	       GtkTreeViewColumn *column)
+{
+	if (curhost != NULL)
+		gdm_chooser_manage (NULL, NULL);
 }
 
 void
@@ -895,6 +1086,7 @@ display_chooser_information (void)
 {
 	GtkWidget *dialog;
 
+	/* How to get HIG compliance? */
 	dialog = gtk_message_dialog_new
 		(GTK_WINDOW (chooser) /* parent */,
 		 GTK_DIALOG_MODAL /* flags */,
@@ -907,8 +1099,10 @@ display_chooser_information (void)
 		   "You can rescan the network for new hosts by clicking "
 		   "\"Refresh\".  When you have selected a host click "
 		   "\"Connect\" to open a session to that machine."));
+	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
 
-	gdm_wm_center_window (GTK_WINDOW (dialog));
+	if (RUNNING_UNDER_GDM)
+		gdm_wm_center_window (GTK_WINDOW (dialog));
 
 	gdm_wm_no_login_focus_push ();
 	gtk_dialog_run (GTK_DIALOG (dialog));
@@ -919,11 +1113,16 @@ display_chooser_information (void)
 static void 
 gdm_chooser_gui_init (void)
 {
+	int width;
+	int height;
+	GtkTreeSelection *selection;
+	GtkTreeViewColumn *column;
+
 	glade_helper_add_glade_directory (GDM_GLADE_DIR);
 	glade_helper_search_gnome_dirs (FALSE);
 
     /* Enable theme */
-    if (GdmGtkRC)
+    if (RUNNING_UNDER_GDM && GdmGtkRC)
 	gtk_rc_parse (GdmGtkRC);
 
     /* Load default host image */
@@ -956,20 +1155,80 @@ gdm_chooser_gui_init (void)
 			       GTK_TYPE_BUTTON);
     status_label = glade_helper_get (chooser_app, "status_label",
 				     GTK_TYPE_LABEL);
+    add_entry = glade_helper_get (chooser_app, "add_entry",
+				  GTK_TYPE_ENTRY);
 
     browser = glade_helper_get (chooser_app, "chooser_iconlist",
-				GNOME_TYPE_ICON_LIST);
-    gnome_icon_list_freeze (GNOME_ICON_LIST (browser));
-    gnome_icon_list_set_separators (GNOME_ICON_LIST (browser), " /-_.");
-    gnome_icon_list_set_icon_width (GNOME_ICON_LIST (browser), GdmIconMaxWidth + 20);
-    gnome_icon_list_set_icon_border (GNOME_ICON_LIST (browser), 2);
-    gnome_icon_list_thaw (GNOME_ICON_LIST (browser));
+				GTK_TYPE_TREE_VIEW);
 
+    gtk_dialog_set_has_separator (GTK_DIALOG (chooser), FALSE);
+
+    gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (browser), TRUE);
+
+    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (browser));
+    gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
+
+    g_signal_connect (selection, "changed",
+		      G_CALLBACK (host_selected),
+		      NULL);
+    g_signal_connect (browser, "row_activated",
+		      G_CALLBACK (row_activated),
+		      NULL);
+
+    browser_model = (GtkTreeModel *)gtk_list_store_new (3,
+							GDK_TYPE_PIXBUF,
+							G_TYPE_STRING,
+							G_TYPE_POINTER);
+    gtk_tree_view_set_model (GTK_TREE_VIEW (browser), browser_model);
+    column = gtk_tree_view_column_new_with_attributes
+	    ("Icon",
+	     gtk_cell_renderer_pixbuf_new (),
+	     "pixbuf", CHOOSER_LIST_ICON_COLUMN,
+	     NULL);
+    gtk_tree_view_append_column (GTK_TREE_VIEW (browser), column);
+
+    column = gtk_tree_view_column_new_with_attributes
+	    ("Hostname",
+	     gtk_cell_renderer_text_new (),
+	     "text", CHOOSER_LIST_LABEL_COLUMN,
+	     NULL);
+    gtk_tree_view_append_column (GTK_TREE_VIEW (browser), column);
+
+    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (browser_model),
+					  CHOOSER_LIST_LABEL_COLUMN,
+					  GTK_SORT_ASCENDING);
+
+
+    if ( ! GdmAllowAdd) {
+	    GtkWidget *w = glade_helper_get (chooser_app, "add_hbox",
+					     GTK_TYPE_HBOX);
+	    gtk_widget_hide (w);
+    }
+
+    gtk_window_get_size (GTK_WINDOW (chooser),
+			 &width, &height);
+    if (RUNNING_UNDER_GDM) {
+	    if (width > gdm_wm_screen.width)
+		    width = gdm_wm_screen.width;
+	    if (height > gdm_wm_screen.height)
+		    height = gdm_wm_screen.height;
+    } else {
+	    if (width > gdk_screen_width ())
+		    width = gdk_screen_width ();
+	    if (height > gdk_screen_height ())
+		    height = gdk_screen_height ();
+    }
     gtk_widget_set_size_request (GTK_WIDGET (chooser), 
-				 (gint) gdk_screen_width() * 0.4, 
-				 (gint) gdk_screen_height() * 0.6);
+				 width, height);
+    gtk_window_set_default_size (GTK_WINDOW (chooser), 
+				 width, height);
+    gtk_window_resize (GTK_WINDOW (chooser), 
+		       width, height);
 
-    gdm_wm_center_window (GTK_WINDOW (chooser));
+    if (RUNNING_UNDER_GDM) {
+	    gtk_widget_show_now (chooser);
+	    gdm_wm_center_window (GTK_WINDOW (chooser));
+    }
 }
 
 
@@ -1081,8 +1340,6 @@ main (int argc, char *argv[])
     int nextopt;
     const char *gdm_version;
 
-    if (g_getenv ("DOING_GDM_DEVELOPMENT") != NULL)
-	    DOING_GDM_DEVELOPMENT = TRUE;
     if (g_getenv ("RUNNING_UNDER_GDM") != NULL)
 	    RUNNING_UNDER_GDM = TRUE;
 
@@ -1108,17 +1365,20 @@ main (int argc, char *argv[])
     }
 
     /* Should be a watch already, but just in case */
-    setup_cursor (GDK_WATCH);
+    if (RUNNING_UNDER_GDM)
+	    setup_cursor (GDK_WATCH);
 
     glade_init();
 
     gdm_chooser_parse_config();
 
-    gdm_wm_screen_init (GdmXineramaScreen);
+    if (RUNNING_UNDER_GDM)
+	    gdm_wm_screen_init (GdmXineramaScreen);
 
     gdm_version = g_getenv ("GDM_VERSION");
 
-    if (gdm_version != NULL &&
+    if (RUNNING_UNDER_GDM &&
+	gdm_version != NULL &&
 	strcmp (gdm_version, VERSION) != 0) {
 	    GtkWidget *dialog;
 
@@ -1126,15 +1386,16 @@ main (int argc, char *argv[])
 
 	    gdm_wm_focus_new_windows (TRUE);
 
-	    dialog = gtk_message_dialog_new (NULL /* parent */,
-					     GTK_DIALOG_MODAL /* flags */,
-					     GTK_MESSAGE_ERROR,
-					     GTK_BUTTONS_OK,
-					     _("The chooser version (%s) does not match the daemon "
-					       "version (%s).\n"
-					       "You have probably just upgraded gdm.\n"
-					       "Please restart the gdm daemon or reboot the computer."),
-					     VERSION, gdm_version);
+	    dialog = ve_hig_dialog_new (NULL /* parent */,
+					GTK_DIALOG_MODAL /* flags */,
+					GTK_MESSAGE_ERROR,
+					GTK_BUTTONS_OK,
+					_("Cannot run chooser"),
+					_("The chooser version (%s) does not match the daemon "
+					  "version (%s).  "
+					  "You have probably just upgraded gdm.  "
+					  "Please restart the gdm daemon or reboot the computer."),
+					VERSION, gdm_version);
 
 	    gtk_widget_show_all (dialog);
 	    gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -1149,7 +1410,8 @@ main (int argc, char *argv[])
     gdm_chooser_gui_init();
     gdm_chooser_signals_init();
 
-    set_background ();
+    if (RUNNING_UNDER_GDM)
+	    set_background ();
 
     hosts = (char **)poptGetArgs (ctx);
     /* when no hosts on the command line, take them from the config */
@@ -1177,10 +1439,12 @@ main (int argc, char *argv[])
     gtk_widget_queue_resize (chooser);
     gtk_widget_show_now (chooser);
 
-    gdm_wm_center_window (GTK_WINDOW (chooser));
+    if (RUNNING_UNDER_GDM)
+	    gdm_wm_center_window (GTK_WINDOW (chooser));
 
-    /* can it ever happen that it'd be NULL here ??? */
-    if (chooser->window != NULL) {
+    if (RUNNING_UNDER_GDM &&
+	/* can it ever happen that it'd be NULL here ??? */
+	chooser->window != NULL) {
 	    gdm_wm_init (GDK_WINDOW_XWINDOW (chooser->window));
 
 	    /* Run the focus, note that this will work no matter what
@@ -1189,7 +1453,11 @@ main (int argc, char *argv[])
 	    gdm_wm_focus_window (GDK_WINDOW_XWINDOW (chooser->window));
     }
 
-    setup_cursor (GDK_LEFT_PTR);
+    if (RUNNING_UNDER_GDM)
+	    setup_cursor (GDK_LEFT_PTR);
+
+    if (GdmAllowAdd)
+	    gtk_widget_grab_focus (add_entry);
 
     gtk_main();
 
